"""
Utils for generating random data and comparing performance
"""
import os
import time
import pickle
import random
from kmeans import kmeans, here
here = here(__file__)

try:
    range = xrange
except NameError:
    pass


def timer():
    start = time.clock()
    return lambda: time.clock() - start


def random_points(n):
    """Returns n random [(x_1, x_2, x_3), w] tuples.

        Constraints:
        0 <= abs(x_N) <= 1<<8
        0 <= w <= 100
        x_N, w are non-negative integers
    """
    rx = lambda: random.randrange(0, 1 << 8)
    rw = lambda: random.randrange(1, 10)
    point = lambda: [(rx(), rx(), rx()), rw()]
    filename = os.path.join(here, "_perf.sample")
    try:
        with open(filename, 'rb') as f:
            points = pickle.load(f)
    except:
        points = []

    diff = n - len(points)
    if diff > 0:
        print("Cache was missing {} points".format(diff))
        t = timer()
        points.extend(point() for _ in range(diff))
        with open(filename, 'wb') as f:
            pickle.dump(points, f)
        elapsed = t()
        print("Added {} points to the cache ({}s)".format(diff, elapsed))

    return ListProxy(points)


class ListProxy(list):
    """Fake sizes by setting length"""
    def __init__(self, data):
        super().__init__(data)
        self.max_length = len(data)
        self.length = self.max_length

    @property
    def length(self):
        return self._length

    @length.setter
    def length(self, n):
        if n > self.max_length:
            raise ValueError(
                "Maximum possible length is " + str(self.max_length))
        self._length = n

    def __len__(self):
        return self.length

    def __iter__(self):
        for i in range(self.length):
            yield self[i]


def main():
    samples = [
        # ~ Common "large" image sizes
        (1920 * 1200,  3),
        (1920 * 1200,  5),
        (1920 * 1200, 15),

        # Personal benchmarks
        (747116, 5),   # Unique pixels in 1920 x 1080 image
        (1095169, 5),  # Unique pixels in 15530 x 8591 image

        # Max unique pixels in rgb 256 image
        (16581375, 5)
    ]

    max_sample = max(sample[0] for sample in samples)
    print("Generating {} random points".format(max_sample))
    t = timer()
    points = random_points(max_sample)
    elapsed = t()
    print("Random points generated ({}s)".format(elapsed))

    def run_test(n, k):
        points.length = n
        t = timer()
        kmeans(points, k)
        elapsed = t()
        return elapsed

    for n, k in samples:
        print("Running test: {} points, {} centers".format(n, k))
        elapsed = run_test(n, k)
        print("N {:9} || K {:3} || E {}".format(n, k, elapsed))

if __name__ == "__main__":
    main()



"""
.. module:: kmeans
   :synopsis: python wrapper for a basic c implementation of the
              k-means algorithm.

.. moduleauthor:: Joe Cross <joe.mcross@gmail.com>

"""
import os
import ctypes
import random
import sysconfig
from ctypes import Structure, c_uint, c_ulong, byref
__all__ = ['kmeans']

# ====================================================
# Hook up c module
HERE = os.path.dirname(os.path.realpath(__file__))
'''http://www.python.org/dev/peps/pep-3149/'''
SUFFIX = sysconfig.get_config_var('SO')
if not SUFFIX:
    SOABI = sysconfig.get_config_var('SOABI')
    SUFFIX = ".{}.so".format(SOABI)

SO_PATH = os.path.join(HERE, 'lib' + SUFFIX)
LIB = ctypes.CDLL(SO_PATH)
# ====================================================


class Point(Structure):
    _fields_ = [
        ('r', c_uint, 8),
        ('g', c_uint, 8),
        ('b', c_uint, 8),
        ('center', c_uint, 8),
        ('count', c_uint, 32)
    ]


class Center(Structure):
    _fields_ = [
        ('r', c_ulong, 64),
        ('g', c_ulong, 64),
        ('b', c_ulong, 64),
        ('count', c_uint, 32)
    ]


def _kmeans(points, k, centers, tolerance, max_iterations):

    if centers:
        if k != len(centers):
            raise ValueError(
                "Provided {} centers but k is {}".format(len(centers), k))
    else:
        centers = random.sample(points, k)

    results = pcenters = (Center * k)()
    for i, center in enumerate(centers):
        (r, g, b), count = center
        pcenters[i] = Center(r=r, g=g, b=b, count=count)
    pcenters = byref(pcenters)

    # Generate points
    n = len(points)
    ppoints = (Point * n)()
    for i, point in enumerate(points):
        (r, g, b), count = point
        ppoints[i] = Point(r=r, g=g, b=b, center=-1, count=count)
    ppoints = byref(ppoints)

    # Compute centers
    LIB.kmeans(ppoints, n, pcenters, k, tolerance, max_iterations)

    # Translate
    return [[result.r, result.g, result.b] for result in results]


def kmeans(points, k, centers=None, tolerance=1, max_iterations=0):
    """Return a list of *k* centers (means).  Initial centers are optional.

    :param points: (values, weight) tuples to find centers of.
            value is a list of integer values.
    :type points: list

    :param k: number of centers to calculate
    :type k: int

    :param centers: initial centers, leave blank to randomly select
    :type centers: list

    :param tolerance: maximum delta to consider the centers stable
    :type tolerance: int

    :param max_iterations: maximum assign/update iterations.  0 to loop until
            tolerance is met.
    :type max_iterations: int

    :rtype: list

    """
    return _kmeans(points=points, k=k, centers=centers,
                   tolerance=tolerance, max_iterations=max_iterations)



from kmeans import kmeans


def test_single_point():
    value = [0, 10, 20]
    points = [
        [value, 1]
    ]
    k = 1
    means = kmeans(points, k)
    assert 1 == len(means)
    assert value == means[0]


def test_single_point_with_guess():
    value = [0, 10, 20]
    guess = [(20, 50, 100), 1]
    points = [
        [value, 1]
    ]
    k = 1
    centers = [
        guess
    ]
    means = kmeans(points, k, centers=centers)
    assert 1 == len(means)
    assert value == means[0]


def test_two_points():
    real_mean = [10, 10, 10]
    points = [
        [(0, 0, 0), 1],
        [(20, 20, 20), 1]
    ]
    k = 1
    means = kmeans(points, k)
    assert 1 == len(means)
    assert real_mean == means[0]


def test_two_points_two_centers():
    values = [
        [0, 10, 20],
        [100, 144, 64]
    ]
    points = [
        [value, 1] for value in values
    ]
    k = 2
    means = kmeans(points, k)
    assert 2 == len(means)
    for value in values:
        assert value in means


def test_two_points_with_weights():
    real_mean = [20, 20, 20]
    points = [
        [(0, 0, 0), 1],
        [(30, 30, 30), 2]
    ]
    k = 1
    means = kmeans(points, k)
    assert 1 == len(means)
    assert real_mean == means[0]



import os
from setuptools import setup, find_packages
from distutils.extension import Extension

HERE = os.path.abspath(os.path.dirname(__file__))
README = open(os.path.join(HERE, 'README.rst')).read()
CKMEANS = Extension(
    'kmeans/lib',
    sources=['kmeans/lib.c'],
    extra_compile_args=['-Wno-error=declaration-after-statement',
                        '-O3', '-std=c99']
)

# The packaging is a little wonky.  I've been fighting with tox to try to
# load the dll from the correct relative path, but
# kmeans.__init__:os.path.dirname(os.path.realpath(__file__)) just
# refuses to point anywhere BUT the .tox path.

# To get over that, I moved everything under the kmeans/ folder and for now,
# if it works that's cool.  It's not clean and I don't like it, but it's also
# 1:30 am

setup(
    name='kmeans',
    version='1.0.1',
    url='http://github.com/numberoverzero/kmeans/',
    license='MIT',
    author='Joe Cross',
    install_requires=[],
    author_email='joe.mcross@gmail.com',
    description='python wrapper for basic c implementation of kmeans',
    long_description=README,
    include_package_data=True,
    platforms='any',
    classifiers=[
        'Programming Language :: Python',
        'Development Status :: 5 - Production/Stable',
        'Natural Language :: English',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Topic :: Software Development :: Libraries :: Python Modules',
        ],
    py_modules=['kmeans'],
    packages=find_packages(exclude=('tests',)),
    ext_modules=[CKMEANS]
)
